from dataclasses import dataclass, field, asdict
from collections.abc import Callable
from abc import ABC, abstractmethod
import pandas as pd
import traceback
import inspect


@dataclass
class SimpleTestExport:
    """A stringified representation of a presyence test. Input and Ouputs are representing `.csv`, function is representing valid python code."""

    title: str
    function: str  # supposed to be python code
    input: str  # supposed to be .csv
    expected_output: str  # supposed to be .csv


@dataclass
class TestResult(ABC):
    successful: bool = field(init=False)
    status: str = field(init=False)

    @abstractmethod
    def __bool__(self):
        """test considered success must return True, failed test must return False."""
        ...

    def __post_init__(self):
        """The children will infer it's property from it's classname and it's boolean value."""
        self.successful = bool(self)
        self.status = self.__class__.__name__


@dataclass
class SimpleTestReport:
    """Contains the test statement as well as its outcome formatted in a way serializable to JSON with dataclasses.asdict() function.

    This is what is passed to the frontend UI.
    """

    test: SimpleTestExport
    result: TestResult


@dataclass
class Success(TestResult):
    """The absence of details means absolute success.

    We can envision a SuccessButWithWarning class that have a warning property, yet evaluate to True, as all success should.
    """

    def __bool__(self):
        return True


@dataclass
class WrongResult(TestResult):
    """A wrong result is caracterized by an output different than what was expected, as well as the assertion error message.

    The pandas DataFrame equality assertion function returns somewhat useful error messaged that would be a shame to not pass forward to the UI.
    """

    testrun_output: str
    assertion_error_message: str

    def __bool__(self):
        return False


@dataclass
class Crash(TestResult):
    """A crash is caracterized by a Traceback of what went wrong."""

    traceback: str

    def __bool__(self):
        return False


@dataclass
class SimpleTest:
    """Define a simple test that takes a function and two dataframes, as well as a title.

    The idea is that if `function(input) === expected_output` then the test is a success, otherwise, it's a failure.

    Currently `presyence` only accepts pd.DataFrame, but accepting pd.Series is doable.
    """

    title: str
    function: Callable
    input: pd.DataFrame  # | pd.Series
    expected_output: pd.DataFrame  # | pd.Series

    def __post_init__(self):
        """
        Enforces correct typing.
        """
        for (name, field_type) in self.__annotations__.items():
            if not isinstance(self.__dict__[name], field_type):
                current_type = type(self.__dict__[name])
                raise TypeError(
                    f"The field `{name}` was assigned by `{current_type}` instead of `{field_type}`"
                )

    def export(self) -> SimpleTestExport:
        """This method transforms every component of a SimpleTest into strings"""
        return SimpleTestExport(
            title=self.title,
            function=inspect.getsource(self.function),
            input=self.input.to_csv(index=False),
            expected_output=self.expected_output.to_csv(index=False),
        )

    def __radd__(self, tests: list):
        if isinstance(tests, list):
            tests.append(self)
            return tests
        else:
            raise TypeError(
                (
                    "presyence SimpleTest can only be added to a list, "
                    f"found:, {type(tests)}"
                )
            )

    def run(self) -> SimpleTestReport:
        """Run the test to create a SimpleTestReport

        This method plan is to:
        1. attempt to run the test
        2. triage exceptions based on (a.) implementation (b.) assertion error (wrong result) (c.) crash during the test
        3. A test result can only be of three valid status (a.) it's a success, (b.) the output is not the one expected, (c.) the function under tests crashed.
        4. Each of this valid status will create a different type of report, requiring different arguments.
        5. Once a test has been ran, we serialize the test to string as well as its outcome. We return both in a SimpleTestReport
        """
        try:
            testrun_output = self.function(self.input)
            if isinstance(self.expected_output, pd.DataFrame):
                assert_same = pd.testing.assert_frame_equal
            elif isinstance(self.expected_output, pd.Series):
                raise NotImplementedError(
                    "pd.Series handling will come in a later version."
                )
                # assert_same = pd.testing.assert_series_equal
            else:
                raise NotImplementedError(
                    "Expected output must be a pandas DataFrame or a Serie"
                )
            if not isinstance(testrun_output, pd.DataFrame | pd.Series):
                raise NotImplementedError
            assert_same(self.expected_output, testrun_output)
        except NotImplementedError:
            # We let NotImplementedError bubble up and crash the program
            raise
        except AssertionError as e:
            result = WrongResult(
                testrun_output=testrun_output.to_csv(index=False),
                assertion_error_message=e.args[
                    0
                ],  # Pandas provide useful error messages so we pass them on to the UI
            )
        except Exception:
            result = Crash(traceback=traceback.format_exc())
        else:
            # absence of Exception means the test was successful
            result = Success()
        finally:
            _test_result = SimpleTestReport(
                test=self.export(),  # the test statement, exported to string
                result=result,  # the test outcome
            )
            return asdict(_test_result)



import typer
from .testrunner.discovery import discover_test_files, extract_tests_from_file
from .reporter.server import run_server_and_inject_payload

app = typer.Typer()


@app.command()
def main():
    """The only command of the package, will run tests and then report.

    This is how it operates:

    1. It starts to look for presyence test files in the current working directory tree, where the `presyence` command is run.

    Note: presyence test files are python files for which the name ends with `spec.sy.py`.
    2. If it hasn't found any matching file, it will exit.
    3. Then it will dynamically load each matching file, attempting to extract tests from it.
    4. Tests must be stored in a variable called TESTS which is a list of `presyence.simpletests.SimpleTest`
    5. If any matching file fails to declare a TESTS variable, the program will abort.
    6. Each lists of SimpleTest will be concatenated to create a big 1D list of SimpleTest.
    7. Then each SimpleTest will be executed by calling the `SimpleTest.run(self)` method, the result of that operation will be a SimpleTestReport, appendended to a big 1D list of test results.
    8. The list of SimpleTestReport will be added to a dictionary, this dictionary will later be converted to JSON and injected into the frontend code by the local server serving the UI.
    9. We launch the local UI server and we provide it with the dictionary payload, this payload will later be unpacked and formated by the Vue.JS frontend code server from this local server.
    10. To quit serving the user need to hit Ctrl-C.
    """
    print("Looking for tests")
    test_files = discover_test_files()
    print(test_files)
    if not test_files:
        print("No test file found. Test file names must end with '.spec.sy.py'")
        exit()
    tests = []
    for test_file in test_files:
        _tests = extract_tests_from_file(test_file)
        if not _tests:
            print(f"No tests found in file {test_file.absolute()}")
            exit()
        tests += _tests
    print("Running tests")
    test_results = []
    for test in tests:
        test_results += [test.run()]
    print("Creating the test report")
    test_report = {"testReport": test_results}
    print("Serving the test report on http://localhost:5000/")
    print("Hit Ctrl-C to quit.")
    run_server_and_inject_payload(test_report)



""" This file is about looking for test files
"""
from pathlib import Path
import importlib.machinery
import importlib.util
import sys
import types


def discover_test_files(pattern="**/*.spec.sy.py", directory=".") -> list[Path]:
    """This will discover file matching the specified pattern in the tree starting at the directory.

    Currently we enforce the pattern (ending in .spec.sy.py) and directory (pwd) but this is purely for simplicity.
    """
    p = Path(directory)
    test_files = p.glob(pattern)
    return list(
        test_files
    )  # return a list of files instead of an iterator for convenience


def extract_tests_from_file(file):
    """Will execute the .py file and get the value of the `TESTS` that should be inside."""
    # get the true absolute name of the file
    file = file.resolve()
    parent, name = file.parent, file.name
    # add parent to path to help with relative imports
    if str(parent) not in sys.path:
        sys.path.append(str(parent))
    # in the next lines we will import the file
    module_name = name
    spec = importlib.util.spec_from_file_location(module_name, file)
    # test_module is the .py containing the tests to run
    test_module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = test_module
    spec.loader.exec_module(test_module)
    # check that a lists named TESTS is present
    tests = test_module.__dict__.get("TESTS", [])
    if not tests:
        raise NotImplementedError(
            "Currently, all `.spec.sy.py` must contain a TESTS list containing SimpleTest"
        )
    return tests



from .access_static import get_website_files, WebsiteStaticFile
from .inject_json_into_js import inject_payload
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.responses import FileResponse, HTMLResponse, Response


def run_server_and_inject_payload(payload):
    """Serve the Vue.JS spa and inject it the reports from the test runner.

    1. Gather each file required for the Vue.JS SPA into a dictionary of `WebsiteStaticFile` objects.
    2. Isolate the files with special roles, here : `assets/worker.js` and `index.js`.
    3. For the worker file, inject the payload inside. We will define its route handler in step 6.
    4. For the index file, redirect every URL that doesn't match to it.
    5. Then create the route handler for the `/` route. (returning `index.html`)
    6. Then for every static file, create a FastAPI route handler that returns the file alongside the correct MIME type.
    7. Then serve the app.
    """
    static_files = get_website_files()
    static_files = {
        f.route: f for f in static_files
    }  # index the static file by their route for easier handling

    # because of the bundling process, the name of the is not exactly `/assets/worker.js`
    worker_route = [k for k in static_files.keys() if k.startswith("/assets/worker")][0]
    worker_file = static_files[worker_route]

    worker_file.content = inject_payload(worker_file.content, payload)

    # we override 404 to redirect to the index.html (which is the SPA entrypoint).
    # the reason is the SPA uses its own router client-side which doesnt match the static files.
    index_file = static_files.pop("/index.html")

    async def catch_all(request, exc):
        return HTMLResponse(index_file.content, status_code=200)

    exceptions = {
        404: catch_all,
    }

    app = FastAPI(exception_handlers=exceptions)

    # define the routes
    @app.get("/")
    def home():
        return HTMLResponse(index_file.content, status_code=200)

    # route factory
    def make_route(static_file: WebsiteStaticFile):
        """Makes FastAPI route from a WebsiteStaticFile object"""
        extension = static_file.route.split(".")[-1]
        common_media_types = {
            "js": "application/javascript",
            "ico": "image/x-icon",
            "css": "text/css",
            "html": "text/html",
        }
        mime_type = common_media_types.get(extension)
        if not mime_type:
            raise NotImplementedError

        @app.get(static_file.route)
        def route():
            return Response(static_file.content, media_type=mime_type)

    for each in static_files.values():
        make_route(each)

    uvicorn.run(app, port=5000, log_level="info")



from dataclasses import dataclass, field
import importlib.resources as resources
from typing import TextIO


@dataclass
class StaticFile:
    """This is meant to represent a static file in this python package.

    Actually it can represents any kind of resource in the package, but right now it's purposed is to be subclassed.
    """

    package_name: str
    resource_name: str
    content: str | bytes = field(init=False)

    def __post_init__(self):
        try:
            self.content = resources.read_text(self.package_name, self.resource_name)
        except UnicodeDecodeError:
            self.content = resources.read_binary(self.package_name, self.resource_name)


@dataclass
class WebsiteStaticFile(StaticFile):
    """Represents a static file used by the Vue.JS SPA. It has a route property allowing to match a URL to that file.

    Currently static files used by the website must be stored in the `www` sub-package (directory) for it to work.
    """

    @property
    def route(self) -> str:
        """URL route corresponding to that static file"""
        parent = "/".join(self.package_name.split("www")[-1].split("."))
        return "/".join([parent, self.resource_name])


def get_website_static_files_from_folder(package_name) -> list[WebsiteStaticFile]:
    """Look for resources accessible from `package_name`

    Python can only load resources in a package if those are at the root of this package.
    For example, in the following structure:
    ```
    mypackage/__init__.py
    mypackage/firstresource.py
    mypackage/subdirectory/__init__.py
    mypackage/subdirectory/secondresource.py
    ```
    One can use `resources.read_test("mypackage", "firstresource.py")` but not `resources.read_test("mypackage/subdirectory", "firstresource.py")`.

    """
    result = []
    folder_files = resources.contents(package_name)
    for file_name in folder_files:
        if resources.is_resource(package_name, file_name):
            if not file_name.startswith("__"):  # ignore files like `__init__.py`
                result.append(WebsiteStaticFile(package_name, file_name))
    return result


def get_website_files(
    packages_containing_static_assets=[
        "presyence.reporter.www",
        "presyence.reporter.www.assets",
    ]
) -> list[WebsiteStaticFile]:
    """This loads resources from all folders containing static assets for the Vue.JS SPA."""
    static_files = []
    for package_name in packages_containing_static_assets:
        static_files += get_website_static_files_from_folder(package_name)
    return static_files



import json


def inject_payload(
    content,
    payload,
    begin_payload="beginPayload:!0,",
    end_payload=",endPayload:!0"
):
    """inject into `content` a json dumps of the `payload` between `begin_payload` and `end_payload`"""
    # locate the start and end of the placeholder
    start_placeholder = content.index(begin_payload) + len(begin_payload)
    end_placeholder = content.index(end_payload)
    # insert a place holder easier to replace instead of the current placeholder
    content = list(content)
    new_placeholder = "REPLACEMEPLEASE"
    content[start_placeholder:end_placeholder] = list(new_placeholder)
    content = "".join(content)
    # transform the payload into json to be inserted in the middle of another json string
    new_payload = json.dumps(
        payload, separators=(",", ":")  # separators with out whitespace to minify it
    )
    trim_first_and_last_character = lambda s: s[1:-1]  # remove `{` and `}`
    new_payload = trim_first_and_last_character(new_payload)

    # inject the stringfied payload in place of the placeholder
    content = content.replace(new_placeholder, new_payload)

    return content
